-------------------------------------------------------------------------------
-- Title      : DIOT WIC gateware
-- Project    : Distributed I/O Tier
-------------------------------------------------------------------------------
-- File       : wic_main.vhd
-- Author     : Greg Daniluk <grzegorz.daniluk@cern.ch>
-- Company    : CERN (BE-CO-HT)
-- Standard   : VHDL
-------------------------------------------------------------------------------
-- Description:
-- This is the main module of Warm Magnet Interlocks Controller gateware. It
-- communicates with nanoFIP FMC, reads status of the current loops, and drives
-- test relays.
-------------------------------------------------------------------------------
--
-- Copyright (c) 2018 CERN
--
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 0.51 (the “License”) (which enables you, at your option,
-- to treat this file as licensed under the Apache License 2.0); you may not
-- use this file except in compliance with the License. You may obtain a copy
-- of the License at http://solderpad.org/licenses/SHL-0.51.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- “AS IS” BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.wic_pkg.all;
use work.diag_pkg.all;

entity wic_main is
  generic (
    g_NUM   : integer := 32);
  port (
    rst_n_i         : in  std_logic;
    clk_i           : in  std_logic;
    
    slots_i   : in  t_wic_slot_in_array(g_NUM-1 downto 0);
    slots_o   : out t_wic_slot_out_array(g_NUM-1 downto 0);

    -- diagnostics data
    diags_i   : in  t_diag_out;

    -- nanoFIP Wishbone
    wb_cyc_o  : out std_logic;
    wb_stb_o  : out std_logic;
    wb_we_o   : out std_logic;
    wb_adr_o  : out std_logic_vector(9 downto 0);
    wb_dat_i  : in  std_logic_vector(7 downto 0);
    wb_dat_o  : out std_logic_vector(7 downto 0);
    wb_ack_i  : in  std_logic;
    rst_o     : out std_logic;
    -- nanoFIP control
    rstin_o    : out std_logic;
    nostat_o   : out std_logic;
    var3_acc_o : out std_logic;
    var3_rdy_i : in  std_logic;
    var2_acc_o : out std_logic;
    var2_rdy_i : in  std_logic;
    var1_acc_o : out std_logic;
    var1_rdy_i : in  std_logic;
    p3_lgth_o  : out std_logic_vector(2 downto 0)
  );
end wic_main;

architecture rtl of wic_main is
  
  -- nanoFIP WB interface constants
  constant c_NF_VAR1_BASE         : unsigned(9 downto 0) := "00" & x"00";
  constant c_NF_VAR1_TYPE         : unsigned(9 downto 0) := "00" & x"00";
  constant c_NF_VAR1_SIZE         : unsigned(9 downto 0) := "00" & x"01";
  constant c_NF_VAR1_DATA_START   : unsigned(9 downto 0) := "00" & x"02";
  constant c_NF_VAR1_DATA_END     : unsigned(9 downto 0) := "00" & x"7F";

  constant c_NF_VAR2_BASE         : unsigned(9 downto 0) := "00" & x"80";
  constant c_NF_VAR2_TYPE         : unsigned(9 downto 0) := "00" & x"80";
  constant c_NF_VAR2_SIZE         : unsigned(9 downto 0) := "00" & x"81";
  constant c_NF_VAR2_DATA_START   : unsigned(9 downto 0) := "00" & x"82";
  constant c_NF_VAR2_DATA_END     : unsigned(9 downto 0) := "00" & x"FF";

  constant c_NF_VAR3_BASE         : unsigned(9 downto 0) := "01" & x"00";
  constant c_NF_VAR3_DATA_START   : unsigned(9 downto 0) := "01" & x"02";
  constant c_NF_VAR3_DATA_END     : unsigned(9 downto 0) := "01" & x"7D";

  constant c_NF_VAR_MAXSIZE       : unsigned(9 downto 0) := to_unsigned(124, 10);
  --------------------------

  type t_state is (IDLE, READ_LEN, READ_VAR1, WRITE_VAR3);

  signal state   : t_state;
  signal adr     : unsigned(9 downto 0);
  signal rd_size : unsigned(7 downto 0);
  signal cnt     : unsigned(7 downto 0);


  -- VAR1 consuming
  signal var1_reg     : t_wic_var; -- max size can be 124 bytes
  signal var1_new     : std_logic; -- triggers update of output signals when new variable is received 
  signal var1_read    : std_logic;

  -- VAR3 producing
  signal var3_rdy_d0   : std_logic;
  signal var3_reg      : t_wic_var;
  signal var3_refresh  : std_logic; -- triggers sending status of all inputs
  signal var3_written  : std_logic;
  signal slots_in_reg  : t_wic_slot_in_array(g_NUM-1 downto 0);
  signal slots_in_synced : t_wic_slot_in_array(g_NUM-1 downto 0);
  signal slots_in_synced_d0 : t_wic_slot_in_array(g_NUM-1 downto 0);
  signal slots_in_tgl  : t_wic_slot_in_array(g_NUM-1 downto 0);
  signal slots_in_tgl_reg : t_wic_slot_in_array(g_NUM-1 downto 0);
  signal inputs_rising   : t_wic_slot_in_array(g_NUM-1 downto 0);
  signal inputs_falling  : t_wic_slot_in_array(g_NUM-1 downto 0);
  signal loops_changed : std_logic;
  signal clr_toggling  : std_logic; -- resets toggling detection when a new cycle starts

begin

  GEN_SLOTS_SYNC: for SL in 0 to g_NUM-1 generate
    GEN_LOOPS_SYNC: for I in 0 to c_IOS_SLOT-1 generate
      cmp_sync_ffs: entity work.gc_sync_ffs
        port map(
          clk_i    => clk_i,
          rst_n_i  => rst_n_i,
          data_i   => slots_i(SL).loops(I),
          synced_o => slots_in_synced(SL).loops(I),
          npulse_o => open,
          ppulse_o => open);
    end generate;
  end generate;

  -----------------------------------------------------------------------------
  --                                                                         --
  --     Main state machine responsible for communication with nanoFIP       --
  --                                                                         --
  -----------------------------------------------------------------------------
  process(clk_i)
  begin
    if rising_edge(clk_i) then
      if rst_n_i = '0' then
        var1_new     <= '0';
        var1_acc_o   <= '0';
        var1_read    <= '0';
        var3_acc_o   <= '0';
        var3_written <= '0';
        wb_cyc_o     <= '0';
        wb_stb_o     <= '0';
        wb_we_o      <= '0';
        wb_dat_o     <= (others=>'0');
        cnt          <= (others=>'0');
        adr          <= (others=>'0');
        var1_reg     <= (others=>(others=>'0'));
        slots_in_reg <= (others=> c_ZERO_SLOT_IN); --f_init_slots;
        slots_in_tgl_reg <= (others=> c_ZERO_SLOT_IN);
        state        <= IDLE;
      else

        case state is
          --------------------------------
          -- wait for new VAR1 consumed or request to produce new VAR3
          when IDLE =>
            var1_new     <= '0';
            var1_acc_o   <= '0';
            var3_acc_o   <= '0';
            var3_written <= '0';
            wb_cyc_o     <= '0';
            wb_stb_o     <= '0';
            wb_we_o      <= '0';
            cnt          <= (others=>'0');
            adr          <= (others=>'0');
            wb_dat_o     <= (others=>'0');
            if var1_rdy_i = '0' then
              var1_read  <= '0';
            end if;

            if var1_rdy_i = '1' and var1_read = '0' then
              -- start reading new VAR1 only if it was not already read
              state <= READ_LEN;
            elsif var3_refresh = '1' and var3_rdy_i = '1' then
              slots_in_reg     <= slots_in_synced;
              slots_in_tgl_reg <= slots_in_tgl;
              state            <= WRITE_VAR3;
            end if;

          --------------------------------
          -- VAR1: first read the size of variable (in bytes)
          when READ_LEN =>
            var1_new     <= '0';
            var1_acc_o   <= '1';
            var3_acc_o   <= '0';
            var3_written <= '0';
            wb_cyc_o     <= '1';
            wb_stb_o     <= '1';
            wb_we_o      <= '0';
            cnt          <= (others=>'0');
            adr          <= c_NF_VAR1_SIZE;
            wb_dat_o     <= (others=>'0');

            if wb_ack_i = '1' then
              wb_cyc_o   <= '0';
              wb_stb_o   <= '0';
              rd_size    <= unsigned(wb_dat_i);
              state      <= READ_VAR1;
            elsif var1_rdy_i = '0' then
              wb_cyc_o   <= '0';
              wb_stb_o   <= '0';
              state      <= IDLE;
            end if;
            
          --------------------------------
          -- VAR1: read the whole variable
          -- this tate handles wishbone bus to read byte-by-byte VAR1 data
          when READ_VAR1 =>
            var1_acc_o   <= '1';
            var3_acc_o   <= '0';
            var3_written <= '0';
            wb_cyc_o     <= '1';
            wb_stb_o     <= '1';
            wb_we_o      <= '0';
            adr          <= c_NF_VAR1_DATA_START + cnt;
            wb_dat_o     <= (others=>'0');
            var1_read    <= '1';  -- mark current VAR1 as processed

            if wb_ack_i = '1' then
              -- read variable byte-by-byte
              var1_reg(to_integer(cnt)) <= wb_dat_i;
              wb_stb_o   <= '0';

              if cnt < rd_size-2 then
                var1_new <= '0';
                state    <= READ_VAR1;
              else
                -- read all bytes, signal new VAR1 reception so that outputs
                -- are updated accordingly
                var1_acc_o <= '0';
                wb_cyc_o   <= '0';
                wb_stb_o   <= '0';
                var1_new   <= '1';
                state      <= IDLE;
              end if;

              cnt <= cnt + 1;
            end if;

          --------------------------------
          -- VAR3: write the whole variable
          -- this state handles wishbone bus to write byte-by-byte VAR3 data
          when WRITE_VAR3 =>
            assert var3_rdy_i = '1' report "NanoFIP var3_rdy became 0 while producing variable" severity failure;

            var1_new     <= '0';
            var1_acc_o   <= '0';
            var3_acc_o   <= '1';
            var3_written <= '1'; -- acknowledge var3_refresh trigger while writing it to the nanoFIP
            wb_cyc_o     <= '1';
            wb_stb_o     <= '1';
            wb_we_o      <= '1';
            adr          <= c_NF_VAR3_DATA_START + cnt;
            wb_dat_o     <= var3_reg(to_integer(cnt));
            if var1_rdy_i = '0' then
              -- in case in the meantime nanoFIP gets new value of VAR1,
              -- be prepared to read it after writing of VAR3 is done
              var1_read  <= '0';
            end if;

            if wb_ack_i = '1' then
              wb_stb_o   <= '0';
              if cnt < c_NF_VAR_MAXSIZE-1 then
                state    <= WRITE_VAR3;
              else
                -- written all bytes, we're done
                var3_acc_o <= '0';
                wb_cyc_o   <= '0';
                wb_stb_o   <= '0';
                state      <= IDLE;
              end if;

              cnt <= cnt + 1;
            end if;

          --------------------------------
          when others =>
            state <= IDLE;
        end case;

      end if;
    end if;
  end process;
  -----------------------------------------------------------------------------
  -----------------------------------------------------------------------------

  --------------------------------
  wb_adr_o  <= std_logic_vector(adr);
  rst_o     <= not rst_n_i;
  rstin_o   <= rst_n_i;
  nostat_o  <= '1';   -- disable nanoFIP status sending
  p3_lgth_o <= "101"; -- length of VAR3: 124 Bytes
  -- not used signals
  var2_acc_o <= '0';

  --------------------------------

  GEN_SLOTS_RELAYS: for I in 0 to g_NUM-1 generate
    process(clk_i)
    begin
      if rising_edge(clk_i) then
        if rst_n_i = '0' then
          slots_o(I) <= c_ZERO_SLOT_OUT;
        elsif var1_new = '1' then
          -- drive relays according to VAR1 just received
          slots_o(I) <= f_parse_var1(var1_reg, I);
        end if;
      end if;
    end process;
  end generate;

  --------------------------------

  -- VAR3: pack loops inputs to VAR3 bytes
  var3_reg <= f_pack_var3(slots_in_synced, slots_in_tgl_reg, g_NUM, diags_i);

  -- detect if state of any loop has changed, it's a trigger to produce new VAR3
  loops_changed <= '1' when(f_detect_change(slots_in_synced, slots_in_reg, g_NUM)) else
                   '0';

  -----------------------------------------------------------------------------
  --                                                                         --
  --     Process monitors when a refresh of VAR3 should be triggered
  --                                                                         --
  -----------------------------------------------------------------------------
  process(clk_i)
  begin
    if rising_edge(clk_i) then
      if rst_n_i = '0' then
        var3_refresh   <= '0';
        var3_rdy_d0    <= '0';
      else
        var3_rdy_d0    <= var3_rdy_i;
        if (var3_rdy_d0 = '0' and var3_rdy_i = '1') or loops_changed = '1' then
          -- force the state machine to write VAR3 if inputs have changed, or
          -- at least after previous VAR3 was read - to show that we're alive.
          var3_refresh <= '1';
        elsif var3_written = '1' then
          var3_refresh <= '0';
        end if;
      end if;
    end if;
  end process;

  -----------------------------------------------------------------------------
  --                                                                         --
  -- Toggling detection ensures that Master is notified through error flags  --
  -- if any of the inputs has changed its state twice (e.g. 0->1 and 1->0)   --
  -- between two VAR3 productions. Therefore toggling registers need to be   --
  -- cleared each time a new VAR3 is sent to the Master (i.e. when var3_rdy_i--
  -- goes down).                                                             --
  --                                                                         --
  -----------------------------------------------------------------------------
  clr_toggling <= var3_rdy_d0 and (not var3_rdy_i); 

  -- for each input on each slot, process detects transitions 0->1 and 1->0
  -- this is used for toggling detection
  GEN_TOGGLE_SLOTS: for SL in 0 to g_NUM-1 generate
    process(clk_i)
      variable i : integer;
    begin
      if rising_edge(clk_i) then
        if rst_n_i = '0' or clr_toggling = '1' then
          inputs_rising(SL)  <= c_ZERO_SLOT_IN;
          inputs_falling(SL) <= c_ZERO_SLOT_IN;
          if rst_n_i = '0' then
            -- clear only on reset
            slots_in_synced_d0(SL) <= c_ZERO_SLOT_IN;
          end if;
        else
          slots_in_synced_d0(SL) <= slots_in_synced(SL);
          for i in 0 to c_IOS_SLOT-1 loop
            if slots_in_synced_d0(SL).loops(i) = '0' and slots_in_synced(SL).loops(i) = '1' then
              -- input "i" changed state: 0->1
              inputs_rising(SL).loops(i) <= '1';
            elsif slots_in_synced_d0(SL).loops(i) = '1' and slots_in_synced(SL).loops(i) = '0' then
              -- input "i" changed state: 1->0
              inputs_falling(SL).loops(i) <= '1';
            end if;
          end loop;
        end if;
      end if;
    end process;

    -- inputs are toggling if they changed state twice between two VAR3 productions
    -- (during one WorldFIP cycle)
    slots_in_tgl(SL).loops <= inputs_rising(SL).loops and inputs_falling(SL).loops;
  end generate;

end rtl;
