/*
 * Copyright (C) 2018 CERN
 * Author: Maciej Suminski <maciej.suminski@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>

#include "libmfipp.h"
#include "nanofip.h"

constexpr int VAR_SIZE = 124;

int main(int argc, char *argv[])
{
    auto mfip = MasterFIP::OpenByLUN(0);

    // Hardware and software config for MasterFIP
    mstrfip_hw_cfg hw_cfg = {0};
    hw_cfg.enable_ext_trig = 0;
    hw_cfg.enable_int_trig = 0;
    hw_cfg.enable_ext_trig_term = 0;
    hw_cfg.turn_around_ustime = 14;         // 14 us corresponds to 1 Mbit/s

    mstrfip_sw_cfg sw_cfg = {0};
    sw_cfg.irq_thread_prio = 50;
    sw_cfg.diag_thread_prio = 20;

    // Cycle configuration (60 ms duration)
    CycleSimple cycle(*mfip, 60000);

    // Agents configuration
    nanoFIP agent1(0x01, VAR_SIZE);
    agent1.AddVariables(cycle.GetPeriodicVarWindow());

    // Variable representing I/O state
    uint8_t prod = 0;

    // Initialization
    for(int i = 0; i < VAR_SIZE; ++i) {
        agent1.SetProdVarData(i, 0);
    }

    // Produced variable handler
    agent1.SetProdVarCb([&](nanoFIP& nf, mstrfip_data* data) {
        ++prod;     // Produce next value in sequence

        for(int i = 0; i < VAR_SIZE; ++i) {
            agent1.SetProdVarData(i, prod);
        }

        std::cout << "variable produced = " << (int) prod << std::endl;

        if (data->status != MSTRFIP_DATA_OK) {
            std::cout << "produced variable cb error " << data->status << std::endl;
        }
    } );

    agent1.SetConsVarCb([&](nanoFIP& nf, mstrfip_data* data) {
        uint8_t cons = nf.GetConsVarData(0);

        for(int i = 1; i < VAR_SIZE; ++i) {
            uint32_t var_data = agent1.GetConsVarData(i);

            if(cons != var_data) {
                std::cout << "consumed variable differs at index " << i
                          << " (" << (int) cons << " != " << (int) var_data << ")" << std::endl;
            }
        }

        std::cout << "variable consumed = " << (int) cons << std::endl;

        if (data->status != MSTRFIP_DATA_OK) {
            std::cout << "produced variable cb error " << data->status << std::endl;
        }
    } );

    mfip->Configure(hw_cfg, sw_cfg);
    mfip->StartCycle(cycle);

    // Keep running until a signal arrives
    WaitForSignal();

    std::cout << "--------------------------------------" << std::endl;

    // Display statistics
    const auto& mfipErrors = mfip->GetErrors();
    std::cout << "MasterFIP interface errors:" << std::endl;
    std::cout << "\tINVALID_ANSWER_ACK: "     << mfipErrors.at(MSTRFIP_INVALID_ANSWER_ACK) << std::endl;
    std::cout << "\tINVALID_BITRATE: "        << mfipErrors.at(MSTRFIP_INVALID_BITRATE) << std::endl;
    std::cout << "\tINVALID_BA_INSTR_COUNT: " << mfipErrors.at(MSTRFIP_INVALID_BA_INSTR_COUNT) << std::endl;
    std::cout << "\tBA_IS_NULL: "             << mfipErrors.at(MSTRFIP_BA_IS_NULL) << std::endl;
    std::cout << "\tBA_MAX_VAR: "             << mfipErrors.at(MSTRFIP_BA_MAX_VAR) << std::endl;
    std::cout << "\tBA_INVALID_VAR_COUNT: "   << mfipErrors.at(MSTRFIP_BA_INVALID_VAR_COUNT) << std::endl;
    std::cout << "\tBA_PER_WIND_CFG_BAD: "    << mfipErrors.at(MSTRFIP_BA_PER_WIND_CFG_BAD) << std::endl;
    std::cout << "\tBA_INVALID_INSTR_PARAM: " << mfipErrors.at(MSTRFIP_BA_INVALID_INSTR_PARAM) << std::endl;
    std::cout << "\tBA_NOT_STOP_STATE: "      << mfipErrors.at(MSTRFIP_BA_NOT_STOP_STATE) << std::endl;
    std::cout << "\tBA_INCONSISTENCY: "       << mfipErrors.at(MSTRFIP_BA_INCONSISTENCY) << std::endl;
    std::cout << "\tBA_WRONG_CYCLE_LENGTH: "  << mfipErrors.at(MSTRFIP_BA_WRONG_CYCLE_LENGTH) << std::endl;
    std::cout << "\tBA_INCONSISTENT_INSTR_SET: " << mfipErrors.at(MSTRFIP_BA_INCONSISTENT_INSTR_SET) << std::endl;
    std::cout << "\tBA_WAIT_WIND_INSTR_MISSING: " << mfipErrors.at(MSTRFIP_BA_WAIT_WIND_INSTR_MISSING) << std::endl;
    std::cout << "\tBA_INVALID_MACROCYCLE: "  << mfipErrors.at(MSTRFIP_BA_INVALID_MACROCYCLE) << std::endl;
    std::cout << "\tINVALID_HW_VERS: "        << mfipErrors.at(MSTRFIP_INVALID_HW_VERS) << std::endl;
    std::cout << "\tTRTL_MSG_COUNT_INCONSISTENCY: " << mfipErrors.at(MSTRFIP_TRTL_MSG_COUNT_INCONSISTENCY) << std::endl;
    std::cout << "\tTRTL_MSG_COUNT_OVERFLOW: "<< mfipErrors.at(MSTRFIP_TRTL_MSG_COUNT_OVERFLOW) << std::endl;
    std::cout << "\tTRTL_MSG_OVERFLOW: "      << mfipErrors.at(MSTRFIP_TRTL_MSG_OVERFLOW) << std::endl;
    std::cout << "\tTRTL_MSG_READ_ERR: "      << mfipErrors.at(MSTRFIP_TRTL_MSG_READ_ERR) << std::endl;
    std::cout << "\tTRTL_MSG_READ_NULL: "     << mfipErrors.at(MSTRFIP_TRTL_MSG_READ_NULL) << std::endl;
    std::cout << "\tPOLL_TIMEOUT: "           << mfipErrors.at(MSTRFIP_POLL_TIMEOUT) << std::endl;
    std::cout << "\tPOLL_ERROR: "             << mfipErrors.at(MSTRFIP_POLL_ERROR) << std::endl;
    std::cout << "\tBA_APER_VAR_NO_CB: "      << mfipErrors.at(MSTRFIP_BA_APER_VAR_NO_CB) << std::endl;
    std::cout << "\tBA_APER_MSG_NO_CB: "      << mfipErrors.at(MSTRFIP_BA_APER_MSG_NO_CB) << std::endl;
    std::cout << "\tBA_PER_VAR_NO_CB: "       << mfipErrors.at(MSTRFIP_BA_PER_VAR_NO_CB) << std::endl;
    std::cout << "\tRESET_MOCKTRTL_ERR: "     << mfipErrors.at(MSTRFIP_RESET_MOCKTRTL_ERR) << std::endl;

    const auto& agentErrors = agent1.GetErrors();
    std::cout << "Agent errors:"  << std::endl;
    std::cout << "\tWrite failed: "   << agentErrors.at(WRITE_FAILED) << std::endl;
    std::cout << "\tMissed data: "    << agentErrors.at(MISSED) << std::endl;
    std::cout << "\tNot refreshed: "  << agentErrors.at(NOT_FRESHED) << std::endl;
    std::cout << "\tNot significant: "<< agentErrors.at(NOT_SIGNIFICANT) << std::endl;
    std::cout << "\tTimeout: "        << agentErrors.at(TIMEOUT) << std::endl;
    std::cout << "\tHW bit errors: "  << agentErrors.at(HW_ERR) << std::endl;
    std::cout << "\tHW wrong number of bytes: " << agentErrors.at(HW_BSZ) << std::endl;
    std::cout << "\tHW bad control byte: " << agentErrors.at(HW_CTRL) << std::endl;
    std::cout << "\tHW bad PDU byte: "<< agentErrors.at(HW_PDU) << std::endl;
    std::cout << "\tUnknown error: "  << agentErrors.at(UNKNOWN) << std::endl;

    std::cout << "Agent frame statistics: " << std::endl;
    for(const auto& stat : agent1.GetDataStats()) {
        std::cout << "\t" << nanoFIP::GetDataStatDesc(stat.first) << ": " << stat.second << std::endl;
    }

    //mfip->StopCycle();

    return 0;
}
